-- JPEG image
-- (c) FSC, 2007

--
-- constants
--
M_SOF0  =   0xffc0
M_SOF1  =   0xffc1
M_SOF2  =   0xffc2
M_SOF3  =   0xffc3

M_DHT   =   0xffc4

M_SOF5  =   0xffc5
M_SOF6  =   0xffc6
M_SOF7  =   0xffc7

M_SOF9  =   0xffc9
M_SOF10 =   0xffca
M_SOF11 =   0xffcb

M_SOF13 =   0xffcd
M_SOF14 =   0xffce
M_SOF15 =   0xffcf

M_RST0  =   0xffd0
M_RST7  =   0xffd7

M_SOI   =   0xffd8
M_EOI   =   0xffd9
M_SOS   =   0xffda
M_DQT   =   0xffdb

M_DRI   =   0xffdd

M_APP0  =   0xffe0
M_APP1  =   0xffe1
M_APP2  =   0xffe2
M_APP3  =   0xffe3
M_APP4  =   0xffe4
M_APP5  =   0xffe5
M_APP6  =   0xffe6
M_APP7  =   0xffe7
M_APP8  =   0xffe8
M_APP9  =   0xffe9
M_APP10 =   0xffea
M_APP11 =   0xffeb
M_APP12 =   0xffec
M_APP13 =   0xffed
M_APP14 =   0xffee
M_APP15 =   0xffef

M_JPG0  =   0xfff0
M_JPG13 =   0xfffd
M_COM   =   0xfffe

--
-- all multi-byte values in JPEG files are in big-endian byte order, all bitfields are in MPEG bit order
--
Motorola()
MPEGBitOrder()

--
-- check length of a marker (which cannot be less than 2 because it includes the size field itself
--
local function CheckMarkerLength(Len)
    if Len < 2 then error("Marker length is less than 2 bytes") end
end

--
-- check that we really got the expected marker
--
local function CheckMarkerId(id, errmsg)
    local MarkerID = PeekUI16()
    if type(id) == "function" then
        if id(MarkerID) then return end
    else
        if MarkerID == id then return end
    end
    error(errmsg)
end

--
-- check both marker ID and its length. Returns size of marker data without the length field
--
local function CheckMarker(id, errmsg)
    if id then CheckMarkerId(id, errmsg) end
    SkipBytes(2)
    local Len = PeekUI16()
    SkipBytes(-2)
    CheckMarkerLength(Len)
    return Len - 2
end

--
-- show marker header
--
local function StartMarker(id, errmsg, comment)
    local Len = CheckMarker(id, errmsg)

    Hex()
    WORD("ID", comment)
    
    HexDec()
    WORD "Size"
    
    CrLf()
    
    return Len
end

--
-- show marker data as byte array
--
local function MarkerData(Len)
    if Len > 0 then
        Hex()
        BYTE("Data", nil, Len)
    end
end

--
-- Unknown markers
--
local function _UnknownMarker()
    NameWidth = 10
    MarkerData(StartMarker(nil, nil, "unknown marker"))
end

UnknownMarker = struct("UnknownMarker", _UnknownMarker)

--
-- M_APP0 marker
--
local function _JFIF()
    BYTE__("App0Type", nil, 5)

    Dec()
    BYTE("Version", "should be 1.02", 2)
    
    local Units = PeekUI8()
    local UnitsComment
    if     Units == 0 then UnitsComment = "no units, X and Y specify the pixel aspect ratio"
    elseif Units == 1 then UnitsComment = "X and Y are dots per inch"
    elseif Units == 2 then UnitsComment = "X and Y are dots per cm"
    end
    BYTE("Units", UnitsComment)
    
    WORD("XDensity", "Horizontal pixel density")
    WORD("YDensity", "Vertical pixel density")
    
    local Xthumb = PeekUI8()
    BYTE("Xthumbnail", "Thumbnail horizontal pixel count")
    local Ythumb = PeekUI8()
    BYTE("Ythumbnail", "Thumbnail vertical pixel count")
    
    if Xthumb ~= 0 and Ythumb ~= 0 then
        Hex()
        BYTE("RGB", "Packed (24-bit) RGB values for the thumbnail pixels, n = Xthumbnail*Ythumbnail", Xthumbnail * Ythumbnail)
    end
end

local function _JFXX(Len)
    local Start = GetFilePosition()
    
    BYTE__("App0Type", nil, 5)
        
    local ExtCode = PeekUI8()
    local ExtCodeComment
    if     ExtCode == 0x10 then ExtCodeComment = "Thumbnail coded using JPEG"
    elseif ExtCode == 0x11 then ExtCodeComment = "Thumbnail stored using 1 byte/pixel"
    elseif ExtCode == 0x13 then ExtCodeComment = "Thumbnail stored using 3 bytes/pixel"
    else                        ExtCodeComment = "unknown thumbnail type" end

    Hex()
    BYTE("extension_code", ExtCodeComment)
    CrLf()
    
    if ExtCode == 0x10 then
        local _BytesRest = BytesRest
        BytesRest = Len - BytesProcessed(Start)    
        JPEG("Thumbnail")
        HideFoldings(true)
        BytesRest = _BytesRest
    
    elseif ExtCode == 0x11 then

        local Xthumbnail = PeekUI8()
        Dec()
        BYTE("Xthumbnail", "Thumbnail horizontal pixel count")
        local Ythumbnail = PeekUI8()
        BYTE("Ythumbnail", "Thumbnail vertical pixel count")
        Hex()
        BYTE("Palette", "24-bit RGB pixel values for the color palette", 768)
        if Xthumbnail ~= 0 and Ythumbnail ~= 0 then BYTE("pixel", "8-bit values for the thumbnail pixels", Xthumbnail * Ythumbnail) end

    elseif ExtCode == 0x13 then
    
        local Xthumbnail = PeekUI8()
        Dec()
        BYTE("Xthumbnail", "Thumbnail horizontal pixel count")
        local Ythumbnail = PeekUI8()
        BYTE("Ythumbnail", "Thumbnail vertical pixel count")
        Hex()
        if Xthumbnail ~= 0 and Ythumbnail ~= 0 then BYTE("RGB", "Packed (24-bit) RGB values for the thumbnail pixels", Xthumbnail * Ythumbnail * 3) end
    
    else BYTE("Unknown", "Unknown JFXX data", Len - 6) end
end

local function _App0()
    NameWidth = 15

    local Len = StartMarker(M_APP0, "M_APP0 marker expected", "M_APP0 - application data marker")

    if Len >= 5 then
        local JFIx = PeekBytes(5).Data
        if JFIx == 'JFIF\0' then return _JFIF(Len)
        elseif JFIx == "JFXX\0" then return _JFXX(Len) end
    end

    MarkerData(Len)
end

App0 = struct("APP0", _App0)

--
-- M_APPx marker
--
local function IsAPPx(id)
    return M_APP0 <= id and id <= M_APP15
end

local function _Appx()
    NameWidth = 15

    local Len = StartMarker(IsAPPx, "M_APPx marker expected", "M_APP" .. PeekUI16() - M_APP0 .. " - application data marker")
    MarkerData(Len)
end

AppX = struct("APPx", _Appx)

local function Appx(id)
    if id == M_APP0 then App0("App0")
    else AppX("App" .. id - M_APP0) end
end

--
-- M_DQT marker
--
local function _QuantTable()
    Dec()
    
    local Pq = ToUI8(PeekBits(4))
    local PqComment
    if     Pq == 0 then PqComment = "8"
    elseif Pq == 1 then PqComment = "16"
    else error("Unknown Pq type in DQT marker") end
    
    BITFIELD("BYTE", "Pq", "Quantization table element precision: " .. PqComment .. "-bit Qk values", 4, ToUI8)
    BITFIELD("BYTE", "Tq", "Quantization table destination identifier", 4, ToUI8)
    
    Hex()
    if Pq == 0 then BYTE("Qk", "Quantization table elements in the zig-zag order, 8-bits/element", 64)
    else WORD("Qk", "Quantization table elements in the zig-zag order, 16-bits/element", 64) end
end

QuantTable = struct("QuantTable", _QuantTable)

local function _Dqt()
    NameWidth = 7
    
    local Len = StartMarker(M_DQT, "M_DQT marker expected", "M_DQT - define quantization table")

    while Len > 0 do
        local Start = GetFilePosition()
        QuantTable("QuantTable")
        Len = Len - BytesProcessed(Start)
        if Len > 0 then CrLf() end
    end
end

Dqt = struct("DQT", _Dqt)

--
-- M_SOS marker
--
local function _ScanPars()
    BYTE("Cs", "Scan component selector")
    BITFIELD("BYTE", "Td", "DC entropy coding table destination selector", 4, ToUI8)
    BITFIELD("BYTE", "Ta", "AC entropy coding table destination selector", 4, ToUI8)
end

ScanPars = struct("ScanPars", _ScanPars)

local function _Sos()
    NameWidth = 5
    local Len = StartMarker(M_SOS, "M_SOS marker expected", "M_SOS - Start of scan")
    
    Dec()
    local Ns = PeekUI8()
    BYTE("Ns", "Number of image components in scan")
    CrLf()
    
    array("struct ScanPars", "ScanPars", ScanPars, Ns)
    
    CrLf()
    BYTE("Ss", "Start of spectral or predictor selection")
    BYTE("Se", "End of spectral selection")
    BITFIELD("BYTE", "Ah", "Successive approximation bit position high", 4, ToUI8)
    BITFIELD("BYTE", "Al", "Successive approximation bit position low or point transform", 4, ToUI8)
end

Sos = struct("SOS", _Sos)

--
-- M_SOFx marker
--
local function _FrameComponent()
    NameWidth = 4
    Dec()
    BYTE("C", "component identifier")

    BITFIELD("BYTE", "H", "Horizontal sampling factor", 4, ToUI8)
    BITFIELD("BYTE", "V", "Vertical sampling factor", 4, ToUI8)
    
    BYTE("Tq", "Quantization table destination selector")
end

FrameComponent = struct("FrameComponent", _FrameComponent)

SOFxComment = 
{
    [M_SOF0]    =   "M_SOF0 - Start Of Frame, non-differential, Huffman coding, Baseline DCT",
    [M_SOF1]    =   "M_SOF1 - Start Of Frame, non-differential, Huffman coding, Extended Sequential DCT",
    [M_SOF2]    =   "M_SOF2 - Start Of Frame, non-differential, Huffman coding, Progressive DCT",
    [M_SOF3]    =   "M_SOF3 - Start Of Frame, non-differential, Huffman coding, Lossless (sequential)",
    [M_SOF5]    =   "M_SOF5 - Start Of Frame, differential, Huffman coding, Differential sequential DCT",
    [M_SOF6]    =   "M_SOF6 - Start Of Frame, differential, Huffman coding, Differential progressive DCT",
    [M_SOF7]    =   "M_SOF7 - Start Of Frame, differential, Huffman coding, Differential lossless (sequential)",
    [M_SOF9]    =   "M_SOF9 - Start Of Frame, non-differential, arithmetic coding, Extended sequential DCT",
    [M_SOF10]   =   "M_SOF10 - Start Of Frame, non-differential, arithmetic coding, Progressive DCT",
    [M_SOF11]   =   "M_SOF11 - Start Of Frame, non-differential, arithmetic coding, Lossless (sequential)",
    [M_SOF13]   =   "M_SOF13 - Start Of Frame, differential, arithmetic coding, Differential sequential DCT",
    [M_SOF14]   =   "M_SOF14 - Start Of Frame, differential, arithmetic coding, Differential progressiveDCT",
    [M_SOF15]   =   "M_SOF15 - Start Of Frame, differential, arithmetic coding, Differential lossless (sequential)",
}

local function IsSOFx(id)
    return SOFxComment[id] ~= nil
end

local function _Sofx()
    NameWidth = 5
    
    StartMarker(IsSOFx, "M_SOFx marker expected", SOFxComment[PeekUI16()])
    
    Dec()
    BYTE("P", "sample precision")
    WORD("Y", "number of lines")
    WORD("X", "number of samples per line")
    
    CrLf()
    local Nf = PeekUI8()
    BYTE("Nf", "number of image components in frame")
    CrLf()
    
    array("struct FrameComponent", "FrameComponent", FrameComponent, Nf)
end

Sofx = struct("SOFx", _Sofx)

--
-- M_JPGx marker
--
local function IsJPGx(id)
    return M_JPG0 <= id and id <= M_JPG13
end

local function _Jpgx()
    local Len = StartMarker(IsJPGx, "M_JPGx marker expected", "M_JPG" .. PeekUI16() - M_JPG0 .. " - reserved for JPEG extensions")
    MarkerData(Len)
end

Jpgx = struct("JPGx", _Jpgx)

--
-- M_DHT marker
--
local function _HuffmanTable()
    local _MaxInlineElements = MaxInlineElements
    MaxInlineElements = 256     -- make all Huffman codes visible
    
    NameWidth = 9
    Dec()
    
    local Tc = ToUI8(PeekBits(4))
    local TcComment = "???"
    if     Tc == 0 then TcComment = "DC table or lossless table"
    elseif Tc == 1 then TcComment = "AC table" end
    BITFIELD("BYTE", "Tc", "Table class: " .. TcComment, 4, ToUI8)
    BITFIELD("BYTE", "Th", "Huffman table destination identifier", 4, ToUI8)
    
    CrLf()
    local Li = PeekBytes(16).Data
    BYTE("Li", "Number of Huffman codes of length i", 16)
    
    Hex()
    for i = 1, 16 do
        local Num = string.byte(Li, i, i)
        if Num ~= 0 then
            BYTE("V" .. i, nil, Num)
        end
    end
    MaxInlineElements = _MaxInlineElements
end

HuffmanTable = struct("HuffmanTable", _HuffmanTable)

local function _Dht()
    local Len = StartMarker(M_DHT, "M_DHT marker expected", "M_DHT - Define Huffman table(s)")

    while Len > 0 do
        local Start = GetFilePosition()
        HuffmanTable "HuffmanTable"
        Len = Len - BytesProcessed(Start)
        if Len > 0 then CrLf() end
    end
end

Dht = struct("DHT", _Dht)

--
-- M_COM marker
--
local function _Com()
    NameWidth = 9
    
    local Len = StartMarker(M_COM, "M_COM marker expected", "Comment");
    
    if IsValidAsciizString(Len) then char__("Data", nil, Len)
    else
        Char()
        char("Data", nil, Len)
    end
end

Com = struct("COM", _Com)

--
-- image data. Returns false if we have eaten the whole data
--
local function ScanData()
    BlockComment("\rCompressed image data\r\r")
    
    local Data = PeekBytes(BytesRest - BytesProcessed(JPEGStart))
    local FindPos = 1

    local FindPattern
    if ProcessRST then
        FindPattern = "\255[^%z]"                                   -- find everything but 0xff00
    else
        FindPattern = "\255[^%z\208\209\210\211\212\213\214\215]"   -- find everything but 0xff00, RST0...RST7
    end
    
    Hex()
    while true do
        local MarkerPos = string.find(Data.Data, FindPattern, FindPos)
        
        if not MarkerPos then
            BYTE("ScanData", nil, BytesRest - BytesProcessed(JPEGStart))
            CrLf()
            BlockComment("\rM_EOI marker not found, image is corrupted\r\r", true)
            return false
        end
        
        local ChunkLen = MarkerPos - FindPos
        
        -- process all image data before the marker
        if ChunkLen > 0 then
            BYTE("ScanData", nil, ChunkLen)
        end
        
        -- check if this is a M_RSTm marker and return otherwise
        local Mrkr = PeekUI16()
        if Mrkr < M_RST0 or Mrkr > M_RST7 then return true end
        
        CrLf()
        local RstNum = Mrkr - M_RST0
        WORD("RST" .. RstNum, "M_RST" .. RstNum .. " - Restart with modulo 8 count " .. RstNum .. "\r\n")
        
        FindPos = FindPos + ChunkLen + 2
    end
end



--
-- SOS and image data
--
local function SosAndImage()
    Sos("Sos")
    CrLf()

    -- make inline arrays smaller and faster
    local _MaxInlineElements = MaxInlineElements
    MaxInlineElements = 8
    local RetVal = ScanData()
    MaxInlineElements = _MaxInlineElements
    
    return RetVal
end

--
-- EOI marker
--
local function Eoi()
    Hex()
    WORD("EOI", "M_EOI - End of image")
    
    -- process any extra data
    local ExtraDataLen = BytesRest - BytesProcessed(JPEGStart)
    if ExtraDataLen > 0 then
        CrLf()
        BlockComment("\rExtraneous data found after the EOI marker\r\r", true)
        BYTE("ExtraData", nil, ExtraDataLen)
    end
end

--
-- DRI marker
--
local function _Dri()
    StartMarker(M_DRI, "M_DRI marker expected", "M_DRI - Define restart interval")
    
    Dec()
    WORD("Ri", "Restart interval - Specifies the number of MCU in the restart interval")
end

Dri = struct("DRI", _Dri)

--
-- the whole JPEG file
--
local function __JPEG()
    NameWidth = 4
    if PeekUI16() ~= M_SOI then BlockComment("\rThis is not a valid JPEG image\r\r", true) return end
    
    Hex()
    WORD("SOI", "M_SOI - Start of image\r\n")
    
    while true do
        if PeekUI8() ~= 0xff then error("Marker (0xff) expected") end
        local NextMarker = PeekUI16()
        
        if     NextMarker == M_SOI  then
        elseif NextMarker == M_DQT  then Dqt("Dqt")
        elseif IsSOFx(NextMarker)   then Sofx("Sof")
        elseif IsAPPx(NextMarker)   then Appx(NextMarker)
        elseif IsJPGx(NextMarker)   then Jpgx("Jpg")
        elseif NextMarker == M_DHT  then Dht("Dht")
        elseif NextMarker == M_COM  then Com("Com")
        elseif NextMarker == M_DRI  then Dri("Dri")
        elseif NextMarker == M_SOS  then if not SosAndImage() then break end
        elseif NextMarker == M_EOI  then Eoi() break
        else                             UnknownMarker("Unknown") end
        CrLf()
    end

    -- don't collapse the top-level JPEG structure
    HideFoldings(false)
end

local function _JPEG()
    -- remember where the whole current JPEG file (this may be a thumbnail!) has started
    local _JPEGStart = JPEGStart
    JPEGStart = GetFilePosition()

    __JPEG()

    -- restore beginning of the file
    JPEGStart = _JPEGStart
end

JPEG = struct("JPEG", _JPEG)

-- the following variables let us process nested JPEG thumbnails the same way as the main JPEG image
BytesRest = BytesTillEOF()
JPEGStart = GetFilePosition()

-- collapse all foldings automatically
HideFoldings(true)

-- set this to false if M_RSTm marker processing is not necessary (a great speed-up!)
ProcessRST = true
